# -*- coding: utf-8 -*-
"""
filtering component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.filtering import FilteringPackage
from pyrin.filtering.manager import FilteringManager


@component(FilteringPackage.COMPONENT_NAME)
class FilteringComponent(Component, FilteringManager):
    """
    filtering component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
validator auto handler module.
"""

from pyrin.validator.interface import AbstractValidatorBase


class AutoValidator(AbstractValidatorBase):
    """
    auto validator class.
    """

    def __init__(self, domain, field, **options):
        """
        initializes an instance of AutoValidator.

        :param type[BaseEntity] domain: the domain in which this validator
                                        must be registered. it must be a
                                        type of a BaseEntity subclass.

        :param InstrumentedAttribute field: validator field. it must be a
                                            column. each validator will be registered
                                            with its field in corresponding domain.

        :keyword AbstractValidatorBase type_validator: the type validator instance.

        :keyword AbstractValidatorBase range_validator: the range validator instance.
                                                        it is either a range validator,
                                                        max validator, min validator or None.

        :keyword AbstractValidatorBase in_validator: the in validator instance.
                                                     it is either an in validator,
                                                     not in validator or None.

        :keyword AbstractValidatorBase custom_validator: a custom validator to extend
                                                         auto validator behavior.
        """

        super().__init__()

        self._name = field.key
        self._domain = domain
        self._type_validator = options.get('type_validator')
        self._range_validator = options.get('range_validator')
        self._in_validator = options.get('in_validator')
        self._custom_validator = options.get('custom_validator')

    def validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        it returns the same or fixed value.

        :param object | list[object] value: value to be validated.

        :keyword bool for_update: specifies that this field is being
                                  validated for update operation.
                                  defaults to False if not provided.

        :keyword bool for_find: specifies that validation is for find operation.
                                defaults to False if not provided.

        :keyword str field_name: a custom field name to be used in validation errors.

        :keyword bool ignore_default: do not consider default value of the column
                                      for nullability validation. defaults to False
                                      if not provided and if the value is None but
                                      the column has default value it will be
                                      considered as valid.

        :raises ValidationError: validation error.

        :returns: object | list[object]
        """

        for_find = options.get('for_find', False)
        fixed_value = value
        if self._type_validator is not None:
            fixed_value = self._type_validator.validate(fixed_value, **options)

        if for_find is False:
            if self._range_validator is not None:
                fixed_value = self._range_validator.validate(fixed_value, **options)

            if self._in_validator is not None:
                fixed_value = self._in_validator.validate(fixed_value, **options)

            if self._custom_validator is not None:
                fixed_value = self._custom_validator.validate(fixed_value, **options)

        return fixed_value

    @property
    def name(self):
        """
        gets the name of this validator.

        :rtype: str
        """

        return self._name

    @property
    def domain(self):
        """
        gets the domain of this validator.

        domain is the type of a BaseEntity subclass that
        this validator validates a value of it.

        :rtype: type[BaseEntity]
        """

        return self._domain

    @property
    def for_find(self):
        """
        gets a value indicating that this validator should only be used on validation for find.

        :rtype: bool
        """

        return False



# -*- coding: utf-8 -*-
"""
filtering services module.
"""

from pyrin.application.services import get_component
from pyrin.filtering import FilteringPackage


def filter(entity, filters, *ignore, **options):
    """
    gets filtering expressions for given entity type based on given filters.

    :param type[pyrin.database.model.base.BaseEntity] entity: entity class type to
                                                              be used for filtering.

    :param dict filters: filters to be applied.

    :param sqlalchemy.orm.attributes.InstrumentedAttribute ignore: columns to be ignored
                                                                   from filtering.

    :keyword bool remove: remove all keys that are applied as filter from filters input.
                          defaults to True if not provided.

    :returns: list of expressions for filtering.
    :rtype: list
    """

    return get_component(FilteringPackage.COMPONENT_NAME).filter(entity, filters,
                                                                 *ignore, **options)



# -*- coding: utf-8 -*-
"""
filtering package.
"""

from pyrin.packaging.base import Package


class FilteringPackage(Package):
    """
    filtering package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'filtering.component'



# -*- coding: utf-8 -*-
"""
filtering manager module.

this manager exposes an interface to perform filtering on entities.
the main usage for this type of filtering is in `find` services to reduce the relevant code.
there are some rules that need to be noticed:

1. filtering on single values (non-string) will be performed using `==`.
2. filtering on single values (string) will be performed using `icontains`.
3. filtering on list values will be performed using `in`.
4. filtering on range values will be performed using `>=` and `<=`.

* note that on primary key columns, no range filtering will be done.
"""

from datetime import datetime

import pyrin.utils.sqlalchemy as sqlalchemy_utils
import pyrin.utilities.range.services as range_services

from pyrin.core.structs import Manager
from pyrin.filtering import FilteringPackage
from pyrin.core.globals import LIST_TYPES


class FilteringManager(Manager):
    """
    filtering manager class.
    """

    package_class = FilteringPackage

    def _get_column_attribute(self, entity, name):
        """
        gets column attribute with given name from given entity type.

        :param type[pyrin.database.model.base.BaseEntity] entity: entity class type to
                                                                  get it column attribute.
        :param str name: column attribute name.

        :rtype: sqlalchemy.orm.attributes.InstrumentedAttribute
        """

        return getattr(entity, name)

    def filter(self, entity, filters, *ignore, **options):
        """
        gets filtering expressions for given entity type based on given filters.

        :param type[pyrin.database.model.base.BaseEntity] entity: entity class type to
                                                                  be used for filtering.

        :param dict filters: filters to be applied.

        :param sqlalchemy.orm.attributes.InstrumentedAttribute ignore: columns to be ignored
                                                                       from filtering.

        :keyword bool remove: remove all keys that are applied as filter from filters input.
                              defaults to True if not provided.

        :returns: list of expressions for filtering.
        :rtype: list
        """

        remove = options.get('remove', True)
        all_columns = entity.primary_key_columns + \
            entity.foreign_key_columns + entity.all_columns

        ignore_names = [item.key for item in ignore]
        expressions = []
        for name in all_columns:
            if name not in ignore_names:
                column = self._get_column_attribute(entity, name)
                collection_type, python_type = column.get_python_type()
                if name in filters:
                    value = filters.get(name)
                    if remove is not False:
                        filters.pop(name, None)

                    if python_type is str and not \
                            isinstance(value, LIST_TYPES) and collection_type is None:
                        expressions.append(column.icontains(value))
                    else:
                        sqlalchemy_utils.add_comparison_clause(expressions, column, value)

                if range_services.is_range_supported_column(column):
                    from_name, to_name = range_services.get_range_filter_names(name)
                    if from_name in filters or to_name in filters:
                        from_value = filters.get(from_name)
                        to_value = filters.get(to_name)
                        if remove is not False:
                            filters.pop(from_name, None)
                            filters.pop(to_name, None)

                        if python_type is datetime:
                            sqlalchemy_utils.add_datetime_range_clause(expressions, column,
                                                                       from_value, to_value)
                        else:
                            sqlalchemy_utils.add_range_clause(expressions, column,
                                                              from_value, to_value)

        return expressions



# -*- coding: utf-8 -*-
"""
utilities range package.
"""

from pyrin.packaging.base import Package


class UtilitiesRangePackage(Package):
    """
    utilities range package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'utilities.range.component'



# -*- coding: utf-8 -*-
"""
swagger package.
"""

from pyrin.packaging.base import Package


class SwaggerPackage(Package):
    """
    swagger package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.configuration']
    COMPONENT_NAME = 'api.swagger.component'
    CONFIG_STORE_NAMES = ['swagger']



# -*- coding: utf-8 -*-
"""
swagger tags protected module.
"""

from pyrin.api.swagger.decorators import tag
from pyrin.api.swagger.tags.base import BaseTag
from pyrin.api.router.handlers.protected import ProtectedRoute


@tag()
class ProtectedTag(BaseTag):
    """
    protected tag class.

    this tag includes all protected routes.
    """

    _name = 'protected'
    _tag = 'Protected'

    def is_accepted(self, rule, method, **options):
        """
        gets a value indicating that this rule is accepted for this tag.

        :param pyrin.api.router.handlers.base.RouteBase rule: rule instance to be processed.
        :param str method: http method name.

        :rtype: bool
        """

        return isinstance(rule, ProtectedRoute)



# -*- coding: utf-8 -*-
"""
swagger component module.
"""

from pyrin.api.swagger import SwaggerPackage
from pyrin.api.swagger.manager import SwaggerManager
from pyrin.application.decorators import component
from pyrin.application.structs import Component


@component(SwaggerPackage.COMPONENT_NAME)
class SwaggerComponent(Component, SwaggerManager):
    """
    swagger component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
swagger tags base module.
"""

from pyrin.api.swagger.interface import AbstractTag
from pyrin.api.swagger.tags.exceptions import TagNameIsRequiredError, TagIsRequiredError


class BaseTag(AbstractTag):
    """
    base tag class.
    """

    def __init__(self, *args, **options):
        """
        initializes an instance of BaseTag.

        :param object args: constructor arguments.

        :raises TagNameIsRequiredError: tag name is required error.
        :raises TagIsRequiredError: tag is required error.
        """

        if self._name in (None, ''):
            raise TagNameIsRequiredError('Name is required for tag [{instance}].'
                                         .format(instance=self))

        if self._tag in (None, ''):
            raise TagIsRequiredError('Tag is required for tag [{instance}].'
                                     .format(instance=self))

        super().__init__()



# -*- coding: utf-8 -*-
"""
range services module.
"""

from pyrin.utilities.range import UtilitiesRangePackage
from pyrin.application.services import get_component


def get_range_filter_names(name):
    """
    gets range filter names for given name.

    it returns a tuple of two items. if name is `created_on`, then
    it returns (`from_created_on`, `to_created_on`) as range filters.

    :param str name: column name.

    :returns: tuple[str lower_bound_name, str upper_bound_name]
    :rtype: tuple[str, str]
    """

    return get_component(UtilitiesRangePackage.COMPONENT_NAME).get_range_filter_names(name)


def is_range_supported_type(type_):
    """
    gets a value indicating that range operations are supported for given type.

    :param type type_: type to be checked for range support.

    :rtype: bool
    """

    return get_component(UtilitiesRangePackage.COMPONENT_NAME).is_range_supported_type(type_)


def is_range_supported_column(column):
    """
    gets a value indicating that range operations are supported for given column.

    columns that are primary keys or their python type is not
    in `RANGE_SUPPORTED_TYPES`, do not support range operations.

    :param sqlalchemy.orm.attributes.InstrumentedAttribute column: column to be checked
                                                                   for range support.

    :rtype: bool
    """

    return get_component(UtilitiesRangePackage.COMPONENT_NAME).is_range_supported_column(column)



# -*- coding: utf-8 -*-
"""
range component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.utilities.range import UtilitiesRangePackage
from pyrin.utilities.range.manager import RangeManager


@component(UtilitiesRangePackage.COMPONENT_NAME)
class RangeComponent(Component, RangeManager):
    """
    range component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
swagger exceptions module.
"""

from pyrin.core.exceptions import CoreException


class SwaggerException(CoreException):
    """
    swagger exception.
    """
    pass


class InvalidTagTypeError(SwaggerException):
    """
    invalid tag type error.
    """
    pass


class DuplicatedTagError(SwaggerException):
    """
    duplicated tag error.
    """
    pass



# -*- coding: utf-8 -*-
"""
range manager module.
"""

from decimal import Decimal
from datetime import datetime, date, time

from pyrin.core.structs import Manager
from pyrin.utilities.range import UtilitiesRangePackage


class RangeManager(Manager):
    """
    range manager class.
    """

    package_class = UtilitiesRangePackage

    # types that support range operations. for example range filtering or range validation.
    RANGE_SUPPORTED_TYPES = (int, float, Decimal, datetime, date, time)

    # keyword prefixes that will be used for range operations.
    # for example range filtering or range validation.
    FROM_KEYWORD_PREFIX = 'from_'
    TO_KEYWORD_PREFIX = 'to_'

    def get_range_filter_names(self, name):
        """
        gets range filter names for given name.

        it returns a tuple of two items. if name is `created_on`, then
        it returns (`from_created_on`, `to_created_on`) as range filters.

        :param str name: column name.

        :returns: tuple[str lower_bound_name, str upper_bound_name]
        :rtype: tuple[str, str]
        """

        from_range = '{prefix}{field}'.format(prefix=self.FROM_KEYWORD_PREFIX, field=name)
        to_range = '{prefix}{field}'.format(prefix=self.TO_KEYWORD_PREFIX, field=name)
        return from_range, to_range

    def is_range_supported_type(self, type_):
        """
        gets a value indicating that range operations are supported for given type.

        :param type type_: type to be checked for range support.

        :rtype: bool
        """

        return type_ in self.RANGE_SUPPORTED_TYPES

    def is_range_supported_column(self, column):
        """
        gets a value indicating that range operations are supported for given column.

        columns that are primary keys or their python type is not
        in `RANGE_SUPPORTED_TYPES`, do not support range operations.

        :param sqlalchemy.orm.attributes.InstrumentedAttribute column: column to be checked
                                                                       for range support.

        :rtype: bool
        """

        __, python_type = column.get_python_type()
        return column.primary_key is not True and self.is_range_supported_type(python_type)



# -*- coding: utf-8 -*-
"""
swagger tags package.
"""

from pyrin.packaging.base import Package


class SwaggerTagsPackage(Package):
    """
    swagger tags package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
swagger services module.
"""

from pyrin.application.services import get_component
from pyrin.api.swagger import SwaggerPackage


def register_tag(instance, **options):
    """
    registers a new tag or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a tag which is already registered.

    :param AbstractTag instance: tag to be registered.
                                 it must be an instance of AbstractTag.

    :keyword bool replace: specifies that if there is another registered
                           tag with the same name, replace it with the
                           new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidTagTypeError: invalid tag type error.
    :raises DuplicatedTagError: duplicated tag error.
    """

    return get_component(SwaggerPackage.COMPONENT_NAME).register_tag(instance, **options)


def get_tags(rule, method, **options):
    """
    gets the valid tags for given rule.

    :param pyrin.api.router.handlers.base.RouteBase rule: rule instance to be processed.
    :param str method: http method name.

    :rtype: list[str]
    """

    return get_component(SwaggerPackage.COMPONENT_NAME).get_tags(rule, method, **options)



# -*- coding: utf-8 -*-
"""
swagger manager module.
"""

import pyrin.configuration.services as config_services

from pyrin.core.structs import Manager, Context
from pyrin.utils.custom_print import print_warning
from pyrin.api.swagger.interface import AbstractTag
from pyrin.api.swagger.structs import ExtendedSwagger
from pyrin.application.services import get_current_app
from pyrin.api.swagger import SwaggerPackage
from pyrin.api.swagger.exceptions import InvalidTagTypeError, DuplicatedTagError


class SwaggerManager(Manager):
    """
    swagger manager class.
    """

    package_class = SwaggerPackage

    def __init__(self):
        """
        initializes an instance of SwaggerManager.
        """

        super().__init__()

        # a dict containing all registered tags in the form of:
        # {str name: AbstractTag instance}
        self._tags = Context()

        self._swagger = None
        if self._is_enabled() is True:
            self._swagger = ExtendedSwagger(get_current_app(),
                                            config=self._get_configs(),
                                            merge=True)

    def _is_enabled(self):
        """
        gets a value indicating that swagger ui is enabled.

        :rtype: bool
        """

        return config_services.get_active('swagger', 'enabled')

    def _is_disabled_tag(self, tag):
        """
        gets a value indicating that given tag is disabled.

        :param str tag: tag name to be checked.

        :rtype: bool
        """

        disabled_tags = config_services.get_active('swagger', 'disabled_tags')
        return tag in disabled_tags

    def _get_custom_tags(self):
        """
        gets a dict of all custom tags.

        :rtype: dict
        """

        return config_services.get_active('swagger', 'custom_tags')

    def _get_configs(self):
        """
        gets configuration from `swagger` config store.

        :rtype: dict
        """

        return config_services.get_active_section('swagger')

    def register_tag(self, instance, **options):
        """
        registers a new tag or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding a tag which is already registered.

        :param AbstractTag instance: tag to be registered.
                                     it must be an instance of AbstractTag.

        :keyword bool replace: specifies that if there is another registered
                               tag with the same name, replace it with the
                               new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidTagTypeError: invalid tag type error.
        :raises DuplicatedTagError: duplicated tag error.
        """

        if not isinstance(instance, AbstractTag):
            raise InvalidTagTypeError('Input parameter [{instance}] is '
                                      'not an instance of [{base}].'
                                      .format(instance=instance,
                                              base=AbstractTag))

        if instance.name in self._tags:
            old_instance = self._tags.get(instance.name)
            replace = options.get('replace', False)
            if replace is not True:
                raise DuplicatedTagError('There is another registered tag '
                                         '[{old}] with name [{name}] but "replace" '
                                         'option is not set, so tag [{instance}] '
                                         'could not be registered.'
                                         .format(old=old_instance,
                                                 name=instance.name,
                                                 instance=instance))

            print_warning('Tag [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance,
                                  new_instance=instance))

        self._tags[instance.name] = instance

    def get_tags(self, rule, method, **options):
        """
        gets the valid tags for given rule.

        :param pyrin.api.router.handlers.base.RouteBase rule: rule instance to be processed.
        :param str method: http method name.

        :rtype: list[str]
        """

        tags = []
        for name, item in self._tags.items():
            if not self._is_disabled_tag(name):
                if item.is_accepted(rule, method):
                    tags.append(item.tag)

        custom_tags = self._get_custom_tags()
        for name, value in custom_tags.items():
            if not self._is_disabled_tag(name):
                if rule.rule.startswith(value):
                    tags.append(name)

        return list(set(tags))



# -*- coding: utf-8 -*-
"""
filtering exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class FilteringException(CoreException):
    """
    filtering exception.
    """
    pass


class FilteringBusinessException(CoreBusinessException, FilteringException):
    """
    filtering business exception.
    """
    pass



# -*- coding: utf-8 -*-
"""
permission package.
"""

from pyrin.packaging.base import Package


class PermissionPackage(Package):
    """
    permission package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'security.permission.component'
